import yaml
import click

from functools import cached_property
from typing import Optional, Dict
from os import path, getenv
from pathlib import Path

from .lib.mql_api import MQLClient, BackendClient
from .lib.shared import MdoModelKey


class CLIContext:  # noqa: D

    API_KEY_CONFIG_KEY = "api_key"
    MQL_OVERRIDE_CONFIG_KEY = "mql_server_url_override"

    def __init__(self) -> None:  # noqa: D
        self.verbose = False
        self.__model_key: Optional[MdoModelKey] = None

        # Check if the proper directory exists, if not create it
        if not path.exists(self.config_dir):
            cofig_dir = Path(self.config_dir)
            cofig_dir.mkdir(parents=True)

    @property
    def config_dir(self) -> str:
        """Transform config dir is ~/.transform."""
        return f"{getenv('HOME')}/.transform"

    @property
    def config_file_path(self) -> str:
        """Config file can be found at ~/.transform/config.yml"""
        return self.config_dir + "/config.yml"

    @property
    def api_key(self) -> Optional[str]:
        """Pull API key from config file by default."""
        return self.__get_config_value(self.API_KEY_CONFIG_KEY)

    @api_key.setter
    def api_key(self, api_key: str) -> None:
        """Set the users' API key and record to the config file."""
        self.__set_config_value(self.API_KEY_CONFIG_KEY, api_key)
        self.identify()
        click.echo("✨ Success! We've set your Transform API key. We're ready to query.")

    def __get_config_value(self, key: str) -> Optional[str]:
        try:
            config_file = open(self.config_file_path, "r")
        except FileNotFoundError:
            return None

        config = yaml.load(config_file, Loader=yaml.Loader)
        if config and key in config and config[key]:
            return config[key]

        return None

    def __set_config_value(self, key: str, value: str) -> None:
        configs: Dict[str, str] = {}
        if path.exists(self.config_file_path):
            with open(self.config_file_path) as f:
                configs = yaml.load(f, Loader=yaml.SafeLoader) or {}

        configs[key] = value

        with open(self.config_file_path, "w") as f:
            yaml.dump(configs, f)

    @cached_property
    def org(self) -> Dict[str, str]:  # noqa: D
        backend_client = BackendClient(self.api_key)
        return backend_client.get_my_org()

    @cached_property
    def user(self) -> Dict[str, str]:  # noqa: D
        backend_client = BackendClient(self.api_key)
        return backend_client.get_me()

    def identify(self) -> None:  # noqa: D
        click.echo(f"You are authenticated as {self.user['userName']} within the {self.org['name']} organization.")

    @property
    def model_key(self) -> MdoModelKey:
        """Defaults to latest Model Key for org if one has not been set."""

        if not self.__model_key:
            backend_client = BackendClient(self.api_key)
            model_key = backend_client.get_latest_model_key()
            self.__model_key = MdoModelKey(
                organization_id=model_key["organizationId"],
                repository=model_key["repository"],
                branch=model_key["branch"],
                commit=model_key["commit"],
            )

        return self.__model_key

    @model_key.setter
    def model_key(self, model_key: MdoModelKey) -> None:
        self.__model_key = model_key

    @property
    def mql_server_url(self) -> str:
        """The URL for the org's MQL server.

        Overrideable via "setup" process for users who need to redirect using a local proxy.
        """
        return self.__get_config_value(self.MQL_OVERRIDE_CONFIG_KEY) or self.org["mqlServerUrl"]

    @mql_server_url.setter
    def mql_server_url(self, mql_server_url: str) -> None:  # noqa: D
        self.__set_config_value(self.MQL_OVERRIDE_CONFIG_KEY, mql_server_url)
        click.echo(f"Success! We've overriden your MQL server to {mql_server_url}")

    @cached_property
    def mql_client(self) -> MQLClient:  # noqa: D
        return MQLClient(self.api_key, self.mql_server_url)
