import click
import ast
import time
import pandas as pd
from typing import Callable, Optional
from .lib.shared import MqlQueryStatus, QueryResult, CacheMode, Constraint

DEFAULT_QUERY_TIMEOUT = 180
POLL_INTERVAL = 1.0


class PythonLiteralOption(click.Option):
    """Click class to decode a string literal into a Python object."""

    def type_cast_value(self, ctx, value):  # type: ignore
        """Cast to Python obj, throws exception if we are unable to cast"""

        # For non-required inputs we receive a None here, and should return it
        if value is None:
            return None
        try:
            return ast.literal_eval(value)
        except Exception:
            raise click.BadParameter(value)


def poll_for_completion(
    id: str,
    get_status: Callable[[str], MqlQueryStatus],
    is_complete: Callable[[MqlQueryStatus], bool],
    is_error: Callable[[MqlQueryStatus], bool],
    verbose: bool = False,
    timeout: int = DEFAULT_QUERY_TIMEOUT,
    poll_interval: float = POLL_INTERVAL,
) -> MqlQueryStatus:
    """Helper function to poll a request for completion.

    Args:
        id: ID of the resource you're waiting for
        get_status: Method to get the resource status
        is_complete: Lambda that will resolve a response on whether we're done polling
        is_error: Lambda that will resolve a response to is_error boolean
        verbose: Verbose mode
        timeout: Defaults to DEFAULT_QUERY_TIMEOUT.
        poll_interval:Defaults to POLL_INTERVAL
    """
    print("Polling for query status…")

    start = time.time()
    while time.time() < start + timeout:
        status = get_status(id)

        if is_error(status):
            # TODO: We need to better parse errors coming from GQL
            raise click.UsageError(f"Query {id} failed with status {status}. See MQL server logs for details.")
        elif is_complete(status):
            print("")  # Serves to flush the last line
            print(f"Query successfully completed after {time.time() - start} seconds")
            return status
        else:
            if verbose:
                print(f"Polling for query completion. Response: {str(status)}")
            else:
                print(".", sep=" ", end="", flush=True)

        time.sleep(poll_interval)

    raise Exception(f"Timeout reached waiting for query {id} to complete after {timeout} seconds.")


def paginate_query_results_to_df(
    id: str,
    get_response: Callable[[str, Optional[int]], QueryResult],
) -> pd.DataFrame:
    """Helper function to poll a request for completion.

    Args:
        id: ID of the resource you're retrieving
        get_response: Method to execute the query itelf

    TODO: Use logging only when verbose
    """
    print("Paging query results…")
    cursor: Optional[int] = 0
    df = pd.DataFrame()

    while cursor is not None:
        queryResult = get_response(id, cursor)
        print(".", sep=" ", end="", flush=True)
        cursor = queryResult.cursor
        df = df.append(queryResult.df, ignore_index=True)

    print("")
    return df


def query_options(function: Callable) -> Callable:
    """Common options for a query, adding support for limit, order, and cache-mode"""
    function = click.option(
        "--cache-mode",
        type=click.Choice(
            list(map(lambda x: x.value, CacheMode)),
            case_sensitive=False,
        ),
        # Cast as CacheMode type
        callback=lambda ctx, param, value: CacheMode(value) if value else None,
    )(function)

    function = click.option(
        "--order",
        required=False,
        cls=PythonLiteralOption,
        help='Metrics or dimensions to order by ("-" in front of a column means descending): syntax is --order \'["-ds"]\'',
    )(function)

    function = click.option(
        "--limit",
        required=False,
        type=int,
        help="Limit the number of rows out: syntax is --limit 100",
    )(function)

    function = click.option(
        "--constraints",
        required=False,
        cls=PythonLiteralOption,
        help='Constraints for the metrics query as Python dict, supports both set and range constraints: \'{"country": {"DE", "FR"}, "ds": ("2020-01-01", "2020-01-30")}\'',
        callback=lambda ctx, param, value: Constraint.from_cli_input(value) if value else None,
    )(function)

    return function


def metrics_and_dimensions_options(function: Callable) -> Callable:
    """TODO: could we validate that these exist using a callback?"""

    function = click.option(
        "--metrics",
        required=True,
        cls=PythonLiteralOption,
        help="Metrics to query for: syntax is --metrics '[\"bookings\"]'",
    )(function)
    function = click.option(
        "--dimensions",
        required=True,
        cls=PythonLiteralOption,
        help="Dimensions to group by: syntax is --dimensions '[\"ds\"]'",
    )(function)
    return function
