import pandas as pd
import base64
from typing import List, Dict, Optional, Tuple
from requests import Response, get


from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport

from ..cli_utils import poll_for_completion, paginate_query_results_to_df
from .shared import Constraint, MdoModelKey, MqlQueryStatus, QueryResult, CacheMode


class BackendClient:  # noqa: D
    def __init__(self, api_key: str) -> None:  # noqa: D
        self._headers = {"Authorization": f"X-Api-Key {api_key}"}
        web_api_transport = RequestsHTTPTransport(
            url="https://api.transformdata.io/graphql", verify=True, retries=2, headers=self._headers
        )
        self._web_api_gql_client = Client(transport=web_api_transport, fetch_schema_from_transport=False)

    def get_latest_model_key(self) -> Dict[str, Dict[str, str]]:
        """Queries for latest model key in Web API GQL"""
        query = gql(
            """
            query LatestModelKeyQuery {
                latestModelKey {
                    organizationId
                    repository
                    branch
                    commit
                }
            }
        """
        )

        return self._web_api_gql_client.execute(query)["latestModelKey"]

    def get_me(self) -> Dict[str, Dict[str, str]]:
        """Queries for "me" (the current user) in Web API GQL"""
        query = gql(
            """
            query MeQuery {
                meDeprecated {
                    id
                    userName
                    email
                    dateCreate
                    dateDeactivate
                    userRoles
                }
            }
        """
        )

        return self._web_api_gql_client.execute(query)["meDeprecated"]

    def get_my_org(self) -> Dict[str, str]:
        """Queries for "my org" (the current user's org) in Web API GQL"""
        query = gql(
            """
            query MyOrgQuery {
                myOrg {
                    name
                    domain
                    dateCreate
                    dateDelete
                    logoUrl
                    primaryConfigRepo
                    primaryConfigBranch
                    mqlServerUrl
                    latestMqlServer {
                    version
                    }
                }
            }
        """
        )

        return self._web_api_gql_client.execute(query)["myOrg"]


class MQLClient:
    """Implements various API interfaces in support of all querying needs"""

    def __init__(self, api_key: str, query_server_url: str) -> None:  # noqa: D
        # https://mql-query.staging.transformdata.io
        self._rest_url = query_server_url
        self._headers = {"Authorization": f"X-Api-Key {api_key}"}
        mql_gql_transport = RequestsHTTPTransport(
            url=f"{query_server_url}/graphql", verify=True, retries=2, headers=self._headers
        )
        self._mql_gql_client = Client(transport=mql_gql_transport, fetch_schema_from_transport=False)

    def materialize(
        self,
        model_key: MdoModelKey,
        table: str,
        metrics: List[str],
        dimensions: List[str],
        constraint: Optional[Constraint] = None,
        cache_mode: Optional[CacheMode] = None,
    ) -> str:
        """Implements Materialize mutation in MQL/GQL"""
        query = gql(
            """
            mutation MaterializeMutation(
                $asTable: String!,
                $groupBy: [String!],
                $metrics: [String!],
                $modelKey: ModelKeyInput,
                $where: ConstraintInput,
                $cacheMode: CacheMode
            ) {
                materialize(
                    asTable: $asTable,
                    groupBy: $groupBy,
                    metrics: $metrics,
                    modelKey: $modelKey,
                    where: $where,
                    cacheMode: $cacheMode
                    ) {
                    materialized
                  }
                }
        """
        )

        params = {
            "modelKey": {
                "organization": model_key.organization_id,
                "repo": model_key.repository,
                "branch": model_key.branch,
                "commit": model_key.commit,
            },
            "asTable": table,
            "groupBy": dimensions,
            "metrics": metrics,
            "where": {
                "constraint": {
                    "constraintType": constraint.constraint_type,
                    "dimensionName": constraint.dimension_name,
                    "values": constraint.values,
                }
                if constraint
                else None
            },
            "cacheMode": (cache_mode or CacheMode.default()).name,
        }

        return self._mql_gql_client.execute(query, variable_values=params)["materialize"]

    def create_query(
        self,
        model_key: MdoModelKey,
        metrics: List[str],
        dimensions: List[str],
        constraints: Optional[List[Constraint]] = None,
        order: Optional[List[str]] = None,
        limit: Optional[int] = None,
        cache_mode: Optional[CacheMode] = None,
    ) -> str:
        """Implements CreateMqlQuery mutation in MQL/GQL"""
        query = gql(
            """
            mutation CreateMqlQueryMutation(
                $modelKey: ModelKeyInput,
                $metrics: [String!],
                $groupBy: [String!],
                $where: ConstraintInput,
                $order: [String!],
                $limit: Int,
                $cacheMode: CacheMode
            ) {
                createMqlQuery(
                input: {
                    modelKey: $modelKey,
                    metrics: $metrics,
                    groupBy: $groupBy,
                    where: $where,
                    order: $order,
                    limit: $limit,
                    cacheMode: $cacheMode
                    postProcessor: "none"
                }
                ) {
                    id
                }
            }
        """
        )

        params = {
            "modelKey": {
                "organization": model_key.organization_id,
                "repo": model_key.repository,
                "branch": model_key.branch,
                "commit": model_key.commit,
            },
            "metrics": metrics,
            "groupBy": dimensions,
            "where": {"And": [c.to_graphql_input() for c in constraints]} if constraints else None,
            "order": order,
            "limit": limit,
            "cacheMode": (cache_mode or CacheMode.default()).name,
        }

        return self._mql_gql_client.execute(query, variable_values=params)["createMqlQuery"]["id"]

    def drop_cache(self) -> bool:
        """Drops the entire MQL cache. Only do this is the cache is somehow corrupt."""
        query = gql(
            """
            mutation {
                dropCache(confirm: "yes") {
                    success
                }
            }
            """
        )

        return bool(self._mql_gql_client.execute(query)["dropCache"]["success"])

    def get_health(self) -> Response:
        """Calls the MQL REST API for Health"""
        return get(f"{self._rest_url}/health", headers=self._headers, timeout=10)

    def get_query_status(self, query_id: str) -> MqlQueryStatus:
        """Retrieves query status given a query_id"""
        status_query = gql(
            """
            query GetMqlQueryResultsStatus($queryId: ID!) {
                mqlQuery(id: $queryId) {
                    status
                }
            }
            """
        )

        return self._mql_gql_client.execute(status_query, variable_values={"queryId": query_id})["mqlQuery"]["status"]

    def get_queries(self, active_only: bool, limit: Optional[int]) -> MqlQueryStatus:
        """Retrieves query status given a query_id"""
        status_query = gql(
            """
            query GetMqlQueries($activeOnly: Boolean, $limit: Int) {
                queries(activeOnly: $activeOnly, limit:$limit) {
                    id,
                    modelKey {
                        branch,
                        commit,
                    },
                    metrics,
                    dimensions,
                    status,
                }
            }
            """
        )

        return self._mql_gql_client.execute(status_query, variable_values={"activeOnly": active_only, "limit": limit})

    def get_logs_by_line(self, query_id: str, from_line: int, max_lines: int = 0) -> Tuple[str, int]:
        """Retrieves logs by log-file line number"""
        log_lines_query = gql(
            """
            query GetLogLines($queryId: ID!, $fromLine: Int, $maxLines: Int) {
                mqlQuery(id: $queryId) {
                    logsByLine(fromLine: $fromLine, maxLines: $maxLines),
                }
            }
            """
        )

        logs = self._mql_gql_client.execute(
            log_lines_query, variable_values={"queryId": query_id, "fromLine": from_line, "maxLines": max_lines}
        )["mqlQuery"]["logsByLine"]
        return logs, len(logs.split("\n"))

    def get_query_page_as_df(self, query_id: str, cursor: Optional[int]) -> QueryResult:
        """Retrieves a single page of a query's results and converts to a DataFrame"""
        results_query = gql(
            """
            query GetMqlQueryResultsTabular($queryId: ID!, $cursor: Int) {
                mqlQuery(id: $queryId) {
                    resultTabular(orient: TABLE, cursor: $cursor) {
                        nextCursor
                        data
                    }
                }
            }
            """
        )
        result = self._mql_gql_client.execute(results_query, variable_values={"queryId": query_id, "cursor": cursor})[
            "mqlQuery"
        ]["resultTabular"]

        return QueryResult(
            df=pd.read_json(
                base64.b64decode(result["data"].encode()),
                orient="table",
            ),
            cursor=result["nextCursor"],
        )

    def poll_for_query_completion(self, query_id: str, timeout: int) -> None:
        """Poll for query completion, displaying some progress indication.

        Raises exception if the query fails or timeout is reached.
        """
        poll_for_completion(
            query_id,
            self.get_query_status,
            lambda status: status == MqlQueryStatus.SUCCESSFUL.name,
            lambda status: (status in [None, MqlQueryStatus.FAILED.name, MqlQueryStatus.UNHANDLED_EXCEPTION.name]),
            False,
            timeout,
        )

    def get_query_dataframe(self, query_id: str, timeout: int) -> pd.DataFrame:
        """Retrieves query results in Pandas DataFrame format

        Includes logic for:
        1. Polling for completion
        2. Paging through results
        3. Converting JSON to a Pandas DataFrame
        """

        self.poll_for_query_completion(query_id, timeout)

        return paginate_query_results_to_df(
            query_id,
            self.get_query_page_as_df,
        )
