import pandas as pd
from typing import List, Optional, Any, Dict
from enum import Enum
from dataclasses import dataclass


class Constraint:  # noqa: D
    CONSTRAINT_TYPE_SET = "SET"
    CONSTRAINT_TYPE_RANGE = "RANGE"

    def __init__(  # noqa: D
        self,
        constraint_type: str,
        dimension_name: str,
        values: Optional[List[str]] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
    ) -> None:
        self.values = None
        self.start = None
        self.end = None

        if constraint_type == self.CONSTRAINT_TYPE_RANGE:
            assert values is None
            assert start is not None
            assert end is not None

            self.start = start
            self.end = end
        elif constraint_type == self.CONSTRAINT_TYPE_SET:
            assert values is not None
            assert start is None
            assert end is None

            self.values = values
        else:
            raise ValueError(f"Invalid Constraint type. Expected 'SET' or 'RANGE', got {constraint_type}")

        self.constraint_type = constraint_type
        self.dimension_name = dimension_name

    def to_graphql_input(self) -> Dict[str, Any]:  # type: ignore [misc]
        """Convert the Constraint into the dict input expected by the createQuery GraphQL mutation"""
        if self.constraint_type == self.CONSTRAINT_TYPE_SET:
            return {
                "constraintType": self.constraint_type,
                "dimensionName": self.dimension_name,
                "values": self.values,
            }
        elif self.constraint_type == self.CONSTRAINT_TYPE_RANGE:
            return {
                "constraintType": self.constraint_type,
                "dimensionName": self.dimension_name,
                "start": self.start,
                "stop": self.end,
            }

    def __str__(self) -> str:
        """Helper function to print the type of constraint being applied"""
        if self.constraint_type == self.CONSTRAINT_TYPE_SET:
            return f"{self.CONSTRAINT_TYPE_SET} constraint on dimension {self.dimension_name} for set {self.values}"
        elif self.constraint_type == self.CONSTRAINT_TYPE_RANGE:
            return f"{self.CONSTRAINT_TYPE_RANGE} constraint on dimension {self.dimension_name} from {self.start} to {self.end}"

    @classmethod
    def from_cli_input(cls, input: Dict[str, Any]) -> List["Constraint"]:  # type: ignore [misc]
        """Create a list of Constraints from a dictionary supplied by the CLI user"""

        if not isinstance(input, dict):
            raise Exception(f"Invalid Constraint input. Expected type dict, got {type(input)}")

        out: List[Constraint] = []
        for dim, values in input.items():
            # Expect a SET of values for a SET contraint
            if isinstance(values, set):
                out.append(cls("SET", dim, list(values)))
            # Expect a two-element tuple of values for a range contraint
            elif isinstance(values, tuple):
                assert len(values) == 2
                out.append(cls("RANGE", dim, None, values[0], values[1]))
            else:
                raise Exception(
                    f"Invalid values input for constraint on dimension {dim}. Expected set or tuple, got {type(values)}"
                )

        return out


"""Copied from tf_thrift"""


class MdoModelKey(object):
    """Copy of tf_thrift MdoModelKey since that should change *very* infrequently and we're attempting to not depend on other transform packages

    Attributes:
     - organization_id
     - repository
     - branch
     - commit

    """

    def __init__(  # noqa: D
        self,
        organization_id=None,
        repository=None,
        branch=None,
        commit=None,
    ) -> None:
        self.organization_id = organization_id
        self.repository = repository
        self.branch = branch
        self.commit = commit


class MqlQueryStatus(Enum):
    """The status of queries submitted for execution in the query manager."""

    # Created and waiting to run.
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    SUCCESSFUL = "SUCCESSFUL"
    # Handled exception prevented a query from being successful.
    FAILED = "FAILED"
    # Unhandled exception prevented a query from being successful.
    UNHANDLED_EXCEPTION = "UNHANDLED_EXCEPTION"
    # The given query ID is not known to exist - possibly expired.
    UNKNOWN = "UNKNOWN"


@dataclass(frozen=True)
class QueryResult:
    """Simple data class for query results with a cursor"""

    df: pd.DataFrame
    cursor: Optional[int] = None


class CacheMode(Enum):
    """Different CacheMode options."""

    READ = "r"  # use cached datasets but do not write new datasets to cache
    READWRITE = "rw"  # use cached datasets and write new datasets to cache
    WRITE = "w"  # ignore cached datatsets and write new datasets to cache
    IGNORE = ""  # ignore cache completely, only use configuration files

    @classmethod
    def from_string(cls, string_value: str) -> "CacheMode":  # noqa: D
        if string_value == "wr":  # convenience rewire
            string_value = "rw"
        return cls(string_value)

    @classmethod
    def default(cls) -> "CacheMode":  # noqa: D
        return CacheMode.READWRITE
