import click
import time

from typing import List, Dict, Optional

from update_checker import UpdateChecker

from .cli_context import CLIContext
from .cli_utils import metrics_and_dimensions_options, query_options, DEFAULT_QUERY_TIMEOUT
from .lib.shared import CacheMode, Constraint, MqlQueryStatus
from . import PACKAGE_NAME, __version__

pass_config = click.make_pass_decorator(CLIContext, ensure=True)


@click.group()
@click.option("-v", "--verbose", is_flag=True)
@pass_config
def cli(config: CLIContext, verbose: bool) -> None:  # noqa: D
    # TODO: fail fast if we don't have an API key and they're not calling "setup"
    #       alternatively can we catch 401 errors from the API and prompt them to setup?
    # TODO: Add support for accepting arbitrary ModelKeys, rather than just "latest"
    config.verbose = verbose

    checker = UpdateChecker()
    result = checker.check(PACKAGE_NAME, __version__)
    # result is None when an update was not found or a failure occured
    if result:
        # Note: As the CLI and API stabilize, we can be a little less aggressive about forcing upgrade
        # For now, we should do so to minimize the likelihood of out of date CLIs causing unexpected behavior
        click.secho(
            "Warning: A new version of the MQL CLI is available.",
            bold=True,
            fg="red",
        )

        click.echo(
            f"Please update to version {result.available_version}, released {result.release_date} by running:\n\t$ pip install --upgrade {PACKAGE_NAME}",
        )
        exit()


@cli.command()
@click.option(
    "-k",
    "--api-key",
    type=str,
    help="Transform API key string",
)
@click.option(
    "-o",
    "--mql-override",
    type=str,
    help="Override the MQL server URL",
)
@pass_config
def setup(config: CLIContext, api_key: str, mql_override: str) -> None:
    """Guides user through CLI setup."""

    if api_key:
        config.api_key = api_key
    # If the user doesn't yet have an API key configured or wants to override it
    elif not config.api_key or click.confirm("We've found an existing API key, do you want to override?", abort=False):
        config.api_key = click.prompt("Please enter your Tranform API key:", type=str, hide_input=True)

    if mql_override:
        config.mql_server_url = mql_override
    elif click.confirm("Override MQL server URL? Required when using a local proxy", default=False, abort=False):
        config.mql_server_url = click.prompt("Enter MQL server URL override", type=str)
        exit()


@cli.command()
@pass_config
def identify(config: CLIContext) -> None:
    """Identify the currently authenticated user."""
    config.identify()


@cli.command()
@pass_config
def drop_cache(config: CLIContext) -> None:
    """Drop the MQL cache. Only necessary if there is evidence cache corruption."""
    resp = config.mql_client.drop_cache()
    if resp:
        click.echo("Successfully dropped MQL cache.")
    else:
        raise click.ClickException("Failed to drop the cache.")


@cli.command()
@pass_config
def health_check(config: CLIContext) -> None:
    """Perform basic HTTP health check against configured MQL server."""
    tic = time.perf_counter()
    resp = config.mql_client.get_health()
    click.echo(
        f"Received HTTP {resp.status_code} code from MQL in {time.perf_counter() - tic:0.4f} seconds. Message: {resp.text}"
    )


@cli.command()
@click.option("--detach", default=False, help="Returns the created query ID to allow for asynchronous querying.")
@click.option(
    "-t",
    "--timeout",
    type=int,
    default=DEFAULT_QUERY_TIMEOUT,
    help="Sets the timeout to wait for query completion. Not applicable in --detach mode.",
)
@metrics_and_dimensions_options
@query_options
@pass_config
def query(
    config: CLIContext,
    detach: bool,
    timeout: int,
    metrics: List[str],
    dimensions: List[str],
    constraints: Optional[List[Constraint]] = None,
    order: Optional[List[str]] = None,
    limit: Optional[int] = None,
    # csv: Optional[str] = None, TODO: Handle csv output
    cache_mode: Optional[CacheMode] = None,
) -> None:
    """Calls library to create a query and assemble a DataFrame from the response"""
    click.echo("Initiating query…")

    if constraints:
        click.echo(", ".join([str(c) for c in constraints]))

    query_id = config.mql_client.create_query(
        model_key=config.model_key,
        metrics=metrics,
        dimensions=dimensions,
        constraints=constraints,
        order=order,
        limit=limit,
        cache_mode=cache_mode,
    )
    if detach:
        click.echo(f"Query ID: {query_id}")
        exit()
    else:
        df = config.mql_client.get_query_dataframe(query_id, timeout)
        # TODO: Offer options for output format
        print(df.to_markdown(index=False))


@cli.command()
@click.option(
    "--table",
    required=True,
    type=str,
    help="Table to materialize",
)
@metrics_and_dimensions_options
@pass_config
def materialize(
    config: CLIContext,
    table: str,
    metrics: List[str],
    dimensions: List[str],
    constraint: Dict[str, str] = None,
    cache_mode: Optional[CacheMode] = None,
) -> None:  # noqa: D
    """WIP"""
    # TODO: I don't know if this is returning the right thing. I'm just getting the string representation of the internal object
    res = config.mql_client.materialize(
        model_key=config.model_key,
        table=table,
        metrics=metrics,
        dimensions=dimensions,
        constraint=constraint,  # type: ignore
        cache_mode=cache_mode,
    )
    print(res)


@cli.command()
@pass_config
@click.option("--query-id", required=True, type=str, help="Query ID to stream logs for")
def stream_query_logs(
    config: CLIContext,
    query_id: str,
) -> None:
    """Retrieve queries from mql server"""
    line_number = 0
    query_status = MqlQueryStatus.UNKNOWN
    while query_status not in [
        MqlQueryStatus.FAILED.name,
        MqlQueryStatus.UNHANDLED_EXCEPTION.name,
        MqlQueryStatus.SUCCESSFUL.name,
    ]:
        logs, line_count = config.mql_client.get_logs_by_line(query_id, line_number)
        line_number += line_count

        query_status = config.mql_client.get_query_status(query_id)

        if line_count > 0:
            print(logs)
        time.sleep(1)


@cli.command()
@click.option("--active-only", type=bool, default=False, help="Return active queries only")
@click.option("--limit", required=False, type=int, help="Limit the number of queries retrieved: syntax is --limit 100")
@pass_config
def list_queries(
    config: CLIContext,
    active_only: bool,
    limit: Optional[int] = None,
) -> None:
    """Retrieve queries from mql server"""
    print(config.mql_client.get_queries(active_only=active_only, limit=limit))


if __name__ == "__main__":
    cli()
